import datetime
import enum
import dataclasses

from injector import inject
from logging import getLogger

from gumo.datastore import EntityKey
from gumo.task_emulator.application.task.repository import TaskProcessRepository
from gumo.task_emulator.domain import GumoTaskProcess
from gumo.task_emulator.domain import TaskState
from gumo.task_emulator.domain import ProcessHistory
from gumo.task_emulator.application.task.encoder import TaskProcessJSONEncoder

logger = getLogger(__name__)


class TaskExecuteState(enum.Enum):
    RUN_AND_SUCCESS = 'run_and_success'
    RUN_AND_FAILED = 'run_and_failed'
    ALREADY_FINISHED = 'already_finished'
    OTHER_RUNNING = 'other_running'


@dataclasses.dataclass(frozen=True)
class TaskExecuteResult:
    state: TaskExecuteState
    task_process: GumoTaskProcess

    def to_json(self) -> dict:
        return {
            'state': self.state.value,
            'taskProcess': TaskProcessJSONEncoder(task_process=self.task_process).to_json(),
        }


class TaskExecuteRunner:
    def run(self, task_process: GumoTaskProcess) -> ProcessHistory:
        raise NotImplementedError()


class TaskExecuteService:
    @inject
    def __init__(
            self,
            task_process_repository: TaskProcessRepository,
            task_execute_runner: TaskExecuteRunner,
    ):
        self._task_process_repository = task_process_repository
        self._task_execute_runner = task_execute_runner

    def execute(self, key: EntityKey) -> TaskExecuteResult:
        task_process = self._task_process_repository.fetch_by_key(key=key)

        if task_process.state.is_finished():
            logger.info(f'TaskProcess.key={key} is already finished. Task execute has been skipped.')
            return TaskExecuteResult(
                state=TaskExecuteState.ALREADY_FINISHED,
                task_process=task_process,
            )

        # if task_process.state.is_processing():
        #     logger.info(f'TaskProcess.key={key} is in progress.')
        #     return TaskExecuteResult(
        #         state=TaskExecuteState.OTHER_RUNNING,
        #         task_process=task_process,
        #     )

        processing = task_process.with_processing()
        self._task_process_repository.save(task_process=processing)

        history = self._task_execute_runner.run(task_process=task_process)
        logger.info(history)

        if history.is_succeeded():
            finished = processing.with_succeeded(history=history)
        else:
            finished = processing.with_failed(history=history)
        self._task_process_repository.save(task_process=finished)

        return TaskExecuteResult(
            state=TaskExecuteState.RUN_AND_SUCCESS,
            task_process=finished,
        )


class TaskMarkAsFailedService:
    @inject
    def __init__(
            self,
            task_process_repository: TaskProcessRepository,
    ):
        self._task_process_repository = task_process_repository

    def execute(self, key: EntityKey) -> GumoTaskProcess:
        task_process = self._task_process_repository.fetch_by_key(key=key)

        if task_process.state.is_finished():
            return task_process

        now = datetime.datetime.utcnow()

        failed = task_process.with_failed(
            history=ProcessHistory(
                started_at=now,
                duration_seconds=0,
                method=task_process.method,
                url=task_process.relative_uri,
                data=None,
            ),
            force_failed=True,
        ).with_state(
            state=TaskState.FAILED,
        )

        self._task_process_repository.save(failed)

        return failed



from logging import getLogger

from injector import inject

from gumo.core.exceptions import ObjectNotoFoundError
from gumo.datastore import datastore_transaction
from gumo.task.domain import GumoTask

from gumo.task_emulator.domain import GumoTaskProcess
from gumo.task_emulator.domain import GumoTaskProcessFactory
from gumo.task_emulator.application.task.repository import TaskRepository
from gumo.task_emulator.application.task.repository import TaskProcessRepository

logger = getLogger(__name__)


class TaskProcessBulkCreateService:
    @inject
    def __init__(
            self,
            task_repository: TaskRepository,
            task_process_repository: TaskProcessRepository,
            task_process_factory: GumoTaskProcessFactory,
    ):
        self._task_repository = task_repository
        self._task_process_repository = task_process_repository
        self._task_process_factory = task_process_factory

    def execute(self) -> dict:
        tasks = self._task_repository.fetch_default_routing_tasks(limit=50)
        converted_tasks = []
        convert_skipped_tasks = []

        for task in tasks:
            logger.debug(f'Convert start Task.key={task.key}')
            task_process = self._build_task_process(task=task)

            if self._convert_task_process(task=task, task_process=task_process):
                converted_tasks.append(task_process)
            else:
                convert_skipped_tasks.append(task_process)

        return {
            'converted_tasks': len(converted_tasks),
            'convert_skipped_tasks': len(convert_skipped_tasks)
        }

    def _build_task_process(self, task: GumoTask) -> GumoTaskProcess:
        return self._task_process_factory.build_from_task(task=task)

    @datastore_transaction()
    def _convert_task_process(self, task: GumoTask, task_process: GumoTaskProcess) -> bool:
        created = self._create_task_process_if_not_exists(task_process=task_process)
        self._task_repository.delete(key=task.key)

        return created

    def _create_task_process_if_not_exists(self, task_process: GumoTaskProcess) -> bool:
        try:
            self._task_process_repository.fetch_by_key(key=task_process.key)
            created = False
        except ObjectNotoFoundError:
            self._task_process_repository.save(task_process=task_process)
            created = True

        return created



from logging import getLogger
import flask.views

from gumo.core.injector import injector

from gumo.task_emulator.application.task.repository import TaskProcessSummaryRepository
from gumo.task_emulator.application.task.repository import TaskProcessRepository

logger = getLogger(__name__)
emulator_ui_blueprint = flask.Blueprint('task-emulator-ui', __name__, template_folder='template')


class QueueUI(flask.views.MethodView):
    _repository = injector.get(TaskProcessSummaryRepository)  # type: TaskProcessSummaryRepository

    def get(self):
        summary = self._repository.fetch_summary()

        return flask.render_template(
            'index.html',
            summary=summary.to_dict(),
        )


class QueueDetailUI(flask.views.MethodView):
    _repository = injector.get(TaskProcessRepository)  # type: TaskProcessRepository

    def get(self, queue_name):
        tasks = self._repository.fetch_tasks(queue_name=queue_name, limit=100)

        return flask.render_template(
            'detail.html',
            queue_name=queue_name,
            tasks=tasks
        )


emulator_ui_blueprint.add_url_rule(
    '/task_emulator/dashboard',
    view_func=QueueUI.as_view(name='dashboard'),
    methods=['GET']
)

emulator_ui_blueprint.add_url_rule(
    '/task_emulator/detail/<queue_name>',
    view_func=QueueDetailUI.as_view(name='dashboard/queue-detail'),
    methods=['GET']
)



from logging import getLogger

import os
import flask.views

from gumo.task_emulator._configuration import configure
from gumo.task_emulator.domain.configuration import TaskEmulatorConfiguration
from gumo.task_emulator.presentation.cli.worker import BackgroundWorker


def task_emulator_flask_blueprints():
    from gumo.task_emulator.presentation.restapi import emulator_api_blueprint
    from gumo.task_emulator.presentation.ui import emulator_ui_blueprint

    return [
        emulator_api_blueprint,
        emulator_ui_blueprint,
    ]


def task_emulator_app():
    flask_app = flask.Flask(__name__)
    flask_app.config['JSON_AS_ASCII'] = False

    for blueprint in task_emulator_flask_blueprints():
        flask_app.register_blueprint(blueprint)

    @flask_app.route('/')
    def root():
        return flask.redirect('/task_emulator/dashboard')

    return flask_app, BackgroundWorker.get_instance()


__all__ = [
    configure.__name__,
    TaskEmulatorConfiguration.__name__,
    BackgroundWorker.__name__,
    task_emulator_app.__name__,
    task_emulator_flask_blueprints.__name__,
]

logger = getLogger('gumo.task_emulator')



import datetime
from logging import getLogger

from injector import inject
from typing import List
from typing import Optional

from gumo.core.exceptions import ObjectNotoFoundError
from gumo.core import EntityKey
from gumo.datastore.infrastructure import DatastoreRepositoryMixin
from gumo.task.domain import GumoTask
from gumo.task.infrastructure.repository.mapper import DatastoreGumoTaskMapper

from gumo.task_emulator.domain import GumoTaskProcess
from gumo.task_emulator.domain import TaskState
from gumo.task_emulator.domain import TaskCount
from gumo.task_emulator.domain import QueueStatus
from gumo.task_emulator.domain import QueueStatusCollection
from gumo.task_emulator.application.task.repository import TaskRepository
from gumo.task_emulator.application.task.repository import TaskProcessRepository
from gumo.task_emulator.application.task.repository import TaskProcessSummaryRepository
from gumo.task_emulator.infrastructure.mapper import DatastoreGumoTaskProcessMapper

logger = getLogger(__name__)


class DatastoreTaskRepository(TaskRepository, DatastoreRepositoryMixin):
    @inject
    def __init__(
            self,
            gumo_task_mapper: DatastoreGumoTaskMapper,
    ):
        super(DatastoreTaskRepository, self).__init__()
        self._task_mapper = gumo_task_mapper

    def _build_query(self):
        return self.datastore_client.query(kind=GumoTask.KIND)

    def _fetch_list(self, query, limit: Optional[int] = None) -> List[GumoTask]:
        tasks = []

        for datastore_entity in query.fetch(limit=limit):
            tasks.append(self._task_mapper.to_entity(
                doc=datastore_entity,
            ))

        return tasks

    def fetch_tasks(self, limit: int = 10) -> List[GumoTask]:
        return self._fetch_list(query=self._build_query(), limit=limit)

    def fetch_default_routing_tasks(self, limit: int = 10) -> List[GumoTask]:
        query = self._build_query()
        query.add_filter('app_engine_routing.service', '=', None)
        query.add_filter('app_engine_routing.version', '=', None)
        query.add_filter('app_engine_routing.instance', '=', None)

        return self._fetch_list(query=query, limit=limit)

    def delete(self, key: EntityKey):
        datastore_key = self.entity_key_mapper.to_datastore_key(entity_key=key)
        self.datastore_client.delete(datastore_key)


class DatastoreTaskProcessRepository(TaskProcessRepository, DatastoreRepositoryMixin):
    @inject
    def __init__(
            self,
            task_process_mapper: DatastoreGumoTaskProcessMapper,
    ):
        super(DatastoreTaskProcessRepository, self).__init__()
        self._task_process_mapper = task_process_mapper

    def _build_query(self):
        return self.datastore_client.query(kind=GumoTaskProcess.KIND)

    def _fetch_query(self, query, limit) -> List[GumoTaskProcess]:
        task_processes = []
        for datastore_entity in query.fetch(limit=(limit or 10)):
            task_processes.append(
                self._task_process_mapper.to_entity(datastore_entity=datastore_entity)
            )

        return task_processes

    def fetch_by_key(self, key: EntityKey) -> GumoTaskProcess:
        datastore_key = self.entity_key_mapper.to_datastore_key(entity_key=key)
        datastore_entity = self.datastore_client.get(key=datastore_key)
        if datastore_entity is None:
            raise ObjectNotoFoundError(f'key={key} is not found.')

        entity = self._task_process_mapper.to_entity(datastore_entity=datastore_entity)
        return entity

    def fetch_tasks_by_state(self, state: TaskState, limit: Optional[int] = None) -> List[GumoTaskProcess]:
        now = datetime.datetime.utcnow().replace(microsecond=0)
        query = self._build_query()
        query.add_filter('state', '=', state.value)
        query.add_filter('run_at', '<=', now)
        query.order = ['run_at']

        task_processes = self._fetch_query(query=query, limit=limit)
        return task_processes

    def fetch_tasks(
            self,
            queue_name: Optional[str] = None,
            limit: Optional[int] = None,
    ) -> List[GumoTaskProcess]:
        query = self._build_query()

        if queue_name:
            query.add_filter('queue_name', '=', queue_name)

        task_processes = self._fetch_query(query=query, limit=limit)
        return task_processes

    def save(self, task_process: GumoTaskProcess):
        datastore_entity = self._task_process_mapper.to_datastore_entity(task_process=task_process)
        logger.debug(f'Datastore Put key={datastore_entity.key}')
        self.datastore_client.put(datastore_entity)

    def cleanup_finished_tasks(self):
        for state in [TaskState.SUCCEEDED.value, TaskState.FAILED.value]:
            query = self._build_query()
            query.add_filter('state', '=', state)
            query.keys_only()

            keys = [task.key for task in query.fetch()]
            logger.info(f'cleanup target keys={len(keys)} items')
            while keys:
                self.datastore_client.delete_multi(keys=keys[:500])
                keys = keys[500:]


class DatastoreTaskProcessSummaryRepository(TaskProcessSummaryRepository, DatastoreRepositoryMixin):
    def _build_query(self):
        return self.datastore_client.query(kind=GumoTaskProcess.KIND)

    def _fetch_queue_names(self) -> List[str]:
        query = self._build_query()
        query.distinct_on = ['queue_name']

        return [t.get('queue_name') for t in query.fetch()]

    def _fetch_task_count_by_status(self, queue_name: str, state: TaskState):
        query = self._build_query()
        query.add_filter('queue_name', '=', queue_name)
        query.add_filter('state', '=', state.value)
        query.keys_only()

        limit = 100
        result = [task for task in query.fetch(limit=limit + 1)]

        if len(result) >= limit + 1:
            return TaskCount(
                count=len(result) - 1,
                has_next=True,
            )
        else:
            return TaskCount(
                count=len(result),
                has_next=False,
            )

    def _fetch_oldest_task_by_queue(self, queue_name: str) -> Optional[datetime.datetime]:
        query = self._build_query()
        query.add_filter('queue_name', '=', queue_name)
        query.add_filter('state', '=', TaskState.QUEUED.value)
        query.order = ['schedule_time']

        for task in query.fetch(limit=1):
            return task.get('schedule_time')

        return None

    def _fetch_queue_status(self, queue_name: str) -> QueueStatus:
        return QueueStatus(
            queue_name=queue_name if queue_name else 'unknown-queue',
            queued_tasks_count=self._fetch_task_count_by_status(queue_name=queue_name, state=TaskState.QUEUED),
            processing_tasks_count=self._fetch_task_count_by_status(queue_name=queue_name, state=TaskState.PROCESSING),
            succeeded_tasks_count=self._fetch_task_count_by_status(queue_name=queue_name, state=TaskState.SUCCEEDED),
            failed_tasks_count=self._fetch_task_count_by_status(queue_name=queue_name, state=TaskState.FAILED),
            oldest_schedule_time=self._fetch_oldest_task_by_queue(queue_name=queue_name),
        )

    def fetch_summary(self) -> QueueStatusCollection:
        queue_names = self._fetch_queue_names()
        queue_statuses = []
        for queue_name in queue_names:
            queue_statuses.append(
                self._fetch_queue_status(queue_name=queue_name)
            )

        return QueueStatusCollection(queue_statuses=queue_statuses)



import dataclasses
import enum
import datetime
import random

from typing import Optional
from typing import List

from gumo.datastore import EntityKey
from gumo.datastore import EntityKeyFactory
from gumo.task.domain import GumoTask


class TaskState(enum.Enum):
    QUEUED = 'queued'
    PROCESSING = 'processing'
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'

    def is_finished(self):
        return self == self.SUCCEEDED or self == self.FAILED

    def is_processing(self):
        return self == self.PROCESSING

    def is_queued(self):
        return self == self.QUEUED

    @classmethod
    def get(cls, value: str):
        try:
            return cls(value)
        except ValueError:
            return cls.QUEUED


@dataclasses.dataclass(frozen=True)
class ProcessRequest:
    method: str
    url: str
    data: Optional[str]
    headers: Optional[dict]

    def data_as_bytes(self) -> Optional[bytes]:
        if self.data is None:
            return
        return self.data.encode('utf-8')


@dataclasses.dataclass(frozen=True)
class ProcessHistory:
    started_at: datetime.datetime
    duration_seconds: float
    method: str
    url: str
    data: Optional[str]
    status_code: Optional[int] = None
    request_headers: Optional[dict] = None
    request_body: Optional[str] = None
    response_headers: Optional[str] = None
    response_body: Optional[str] = None
    error_message: Optional[str] = None

    def is_succeeded(self) -> bool:
        if self.status_code is None:
            return False
        if self.error_message is not None:
            return False

        return 200 <= self.status_code < 400


@dataclasses.dataclass(frozen=True)
class GumoTaskProcess:
    KIND = 'GumoTaskProcess'
    MAX_RETRY_COUNT = 15

    key: EntityKey
    relative_uri: str
    method: str = 'POST'
    payload: Optional[dict] = dataclasses.field(default_factory=dict)
    headers: Optional[dict] = dataclasses.field(default_factory=dict)
    schedule_time: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.utcnow)
    created_at: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.utcnow)
    updated_at: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.utcnow)
    queue_name: Optional[str] = None
    state: TaskState = TaskState.QUEUED
    attempts: int = 0
    last_run_at: Optional[datetime.datetime] = None
    run_at: Optional[datetime.datetime] = None
    locked_at: Optional[datetime.datetime] = None
    succeeded_at: Optional[datetime.datetime] = None
    failed_at: Optional[datetime.datetime] = None
    histories: List[ProcessHistory] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        if self.key.kind() != self.KIND:
            raise ValueError(f'key KIND must be a {self.KIND}, but got: {self.key.kind()}')

    def _clone(self, **changes):
        """
        一部の attributes を変更した、自身のオブジェクトの複製を返す

        :param changes: 変更する attributes
        :return: 一部の attributes が変更された、新しい TextExtractionJob オブジェクト
        :rtype: GumoTaskProcess
        """
        return dataclasses.replace(self, **changes)

    def with_history(self, history: ProcessHistory):
        """
        :rtype: GumoTaskProcess
        """
        histories = self.histories
        histories.append(history)

        return self._clone(histories=histories)

    def with_state(self, state: TaskState):
        """
        :rtype: GumoTaskProcess
        """
        return self._clone(state=state)

    def with_processing(self):
        return self.with_state(
            state=TaskState.PROCESSING
        )._clone(
            locked_at=datetime.datetime.utcnow(),
        )

    def with_succeeded(self, history: ProcessHistory):
        return self.with_state(
            state=TaskState.SUCCEEDED
        ).with_history(
            history=history
        )._clone(
            succeeded_at=datetime.datetime.utcnow(),
            locked_at=None,
        )

    def reach_max_retries(self):
        return self.attempts > self.MAX_RETRY_COUNT

    def with_failed(self, history: ProcessHistory, force_failed: bool = False):
        if self.reach_max_retries() or force_failed:
            return self._with_failed_permanent(history=history)
        else:
            return self._with_failed_to_retry(history=history)

    def _with_failed_to_retry(self, history: ProcessHistory):
        return self.with_state(
            state=TaskState.QUEUED
        ).with_history(
            history=history
        )._clone(
            attempts=self.attempts + 1,
            last_run_at=history.started_at,
            run_at=self._rescheduled_at(time=self.run_at, attempts_count=self.attempts + 1),
            locked_at=None,
        )

    def _with_failed_permanent(self, history: ProcessHistory):
        return self.with_state(
            state=TaskState.FAILED
        ).with_history(
            history=history
        )._clone(
            attempts=self.attempts + 1,
            last_run_at=history.started_at,
            failed_at=datetime.datetime.utcnow(),
            locked_at=None,
        )

    def _rescheduled_at(self, time: datetime.datetime, attempts_count) -> datetime.datetime:
        base_offset = 1
        random_offset = random.random() * 5
        exponential_offset = pow(attempts_count, 4)
        delta = datetime.timedelta(seconds=base_offset + random_offset + exponential_offset)

        return time + delta


class GumoTaskProcessFactory:
    def build_from_task(self, task: GumoTask) -> GumoTaskProcess:
        now = datetime.datetime.utcnow().replace(microsecond=0)
        run_at = task.schedule_time if task.schedule_time else now

        return GumoTaskProcess(
            key=EntityKeyFactory().build(kind=GumoTaskProcess.KIND, name=task.key.name()),
            relative_uri=task.relative_uri,
            method=task.method,
            payload=task.payload,
            headers=task.headers,
            schedule_time=task.schedule_time,
            created_at=task.created_at,
            updated_at=now,
            queue_name=task.queue_name,
            state=TaskState.QUEUED,
            attempts=0,
            last_run_at=None,
            run_at=run_at,
            failed_at=None,
            histories=[],
        )

    def build_with_new_history(
            self,
            task_process: GumoTaskProcess,
            history: ProcessHistory
    ) -> GumoTaskProcess:
        histories = task_process.histories
        histories.append(history)
        return dataclasses.replace(task_process, histories=histories)


@dataclasses.dataclass(frozen=True)
class TaskCount:
    count: int
    has_next: bool

    def __str__(self):
        if self.has_next:
            return f'{self.count}+'
        else:
            return f'{self.count}'


@dataclasses.dataclass(frozen=True)
class QueueStatus:
    queue_name: str
    queued_tasks_count: TaskCount
    processing_tasks_count: TaskCount
    succeeded_tasks_count: TaskCount
    failed_tasks_count: TaskCount
    oldest_schedule_time: Optional[datetime.datetime]

    def to_dict(self) -> dict:
        return {
            'queue_name': self.queue_name,
            'queued_tasks_count': str(self.queued_tasks_count),
            'processing_tasks_count': str(self.processing_tasks_count),
            'succeeded_tasks_count': str(self.succeeded_tasks_count),
            'failed_tasks_count': str(self.failed_tasks_count),
            'oldest_schedule_time': self.oldest_schedule_time.isoformat() if self.oldest_schedule_time else None,
        }


@dataclasses.dataclass(frozen=True)
class QueueStatusCollection:
    queue_statuses: List[QueueStatus]

    def to_dict(self) -> List[dict]:
        return [
            q.to_dict() for q in self.queue_statuses
        ]



import json
from injector import inject

from gumo.datastore.infrastructure import DatastoreEntity
from gumo.datastore.infrastructure import EntityKeyMapper
from gumo.task_emulator.domain import GumoTaskProcess
from gumo.task_emulator.domain import ProcessHistory
from gumo.task_emulator.domain import TaskState


class DatastoreGumoProcessHistoryMapper:
    def to_datastore_entity(self, process_history: ProcessHistory) -> DatastoreEntity:
        doc = DatastoreEntity(exclude_from_indexes=[
            "data", "request_headers", "request_body", "response_headers", "response_body", "error_message"
        ])
        doc.update({
            'started_at': process_history.started_at,
            'duration_seconds': process_history.duration_seconds,
            'method': process_history.method,
            'url': process_history.url,
            'data': json.dumps(process_history.data),
            'status_code': process_history.status_code,
            'request_headers': json.dumps(process_history.request_headers),
            'request_body': process_history.request_body,
            'response_headers': json.dumps(process_history.response_headers),
            'response_body': process_history.response_body,
            'error_message': process_history.error_message,
        })
        return doc

    def to_entity(self, doc: dict) -> ProcessHistory:
        return ProcessHistory(
            started_at=doc.get('started_at'),
            duration_seconds=doc.get('duration_seconds'),
            method=doc.get('method'),
            url=doc.get('url'),
            data=json.loads(doc.get('data')),
            status_code=doc.get('status_code'),
            request_headers=json.loads(doc.get('request_headers')),
            request_body=doc.get('request_body'),
            response_headers=json.loads(doc.get('response_headers')),
            response_body=doc.get('response_body'),
            error_message=doc.get('error_message'),
        )


class DatastoreGumoTaskProcessMapper:
    @inject
    def __init__(
            self,
            entity_key_mapper: EntityKeyMapper,
            process_history_mapper: DatastoreGumoProcessHistoryMapper,
    ):
        self._entity_key_mapper = entity_key_mapper
        self._process_history_mapper = process_history_mapper

    def to_datastore_entity(self, task_process: GumoTaskProcess) -> DatastoreEntity:
        entity = DatastoreEntity(
            key=self._entity_key_mapper.to_datastore_key(entity_key=task_process.key),
            exclude_from_indexes=["payload"]
        )
        entity.update({
            'relative_uri': task_process.relative_uri,
            'method': task_process.method,
            'payload': task_process.payload,
            'headers': task_process.headers,
            'schedule_time': task_process.schedule_time,
            'created_at': task_process.created_at,
            'updated_at': task_process.updated_at,
            'queue_name': task_process.queue_name,
            'state': task_process.state.value,
            'attempts': task_process.attempts,
            'last_run_at': task_process.last_run_at,
            'run_at': task_process.run_at,
            'locked_at': task_process.locked_at,
            'succeeded_at': task_process.succeeded_at,
            'failed_at': task_process.failed_at,
            'histories': [
                self._process_history_mapper.to_datastore_entity(history) for history in task_process.histories
            ],
        })
        return entity

    def to_entity(self, datastore_entity: DatastoreEntity) -> GumoTaskProcess:
        key = self._entity_key_mapper.to_entity_key(datastore_entity.key)
        return GumoTaskProcess(
            key=key,
            relative_uri=datastore_entity.get('relative_uri'),
            method=datastore_entity.get('method'),
            payload=datastore_entity.get('payload'),
            headers=datastore_entity.get('headers'),
            schedule_time=datastore_entity.get('schedule_time'),
            created_at=datastore_entity.get('created_at'),
            updated_at=datastore_entity.get('updated_at'),
            queue_name=datastore_entity.get('queue_name'),
            state=TaskState.get(datastore_entity.get('state')),
            attempts=datastore_entity.get('attempts'),
            last_run_at=datastore_entity.get('last_run_at'),
            run_at=datastore_entity.get('run_at'),
            locked_at=datastore_entity.get('locked_at'),
            succeeded_at=datastore_entity.get('succeeded_at'),
            failed_at=datastore_entity.get('failed_at'),
            histories=[
                self._process_history_mapper.to_entity(history) for history in datastore_entity.get('histories')
            ],
        )



import threading
import time
import logging

from typing import Optional
from gumo.core.injector import injector

from gumo.task_emulator.application import TaskExecuteService
from gumo.task_emulator.application.task import TaskProcessBulkCreateService
from gumo.task_emulator.application.task.repository import TaskProcessRepository
from gumo.task_emulator.domain import TaskState


logger = logging.getLogger(__name__)


class RoutineWorkerStop(RuntimeError):
    pass


class RoutineWorker:
    def __init__(self, interval_seconds: int = 1):
        self._stop_event = threading.Event()
        self._thread = threading.Thread(target=self.loop)
        self._interval_seconds = interval_seconds
        self._started = False

        self._worker_name = f'RoutineWorker({self.__class__.__name__})'

    def start(self, daemon: bool = False, interval_seconds: Optional[int] = None):
        if self._started:
            logger.debug(f'{self._worker_name} is already started.')
            return

        if interval_seconds:
            self._interval_seconds = interval_seconds

        self._stop_event.clear()
        self._thread.setDaemon(daemon)
        self._thread.start()
        logger.debug(f'{self._worker_name} is starting with interval = {self._interval_seconds}sec.')

    def stop(self, wait=True):
        self._stop_event.set()
        if wait:
            self._thread.join()

    def loop(self):
        try:
            while True:
                self.run()

                for i in range(self._interval_seconds):
                    time.sleep(1)
                    self.check_stop()
        except RoutineWorkerStop:
            logger.debug(f'{self._worker_name} is stopped by signals.')

    def check_stop(self):
        if self._stop_event.is_set():
            raise RoutineWorkerStop('Received stop event.')

    def run(self):
        raise NotImplementedError()


class NoopTask(RoutineWorker):
    def run(self):
        time.sleep(1)


class WatchNewTask(RoutineWorker):
    def run(self):
        service = injector.get(TaskProcessBulkCreateService)  # type: TaskProcessBulkCreateService
        service.execute()


class ExecuteTask(RoutineWorker):
    def run(self):
        repository = injector.get(TaskProcessRepository)  # type: TaskProcessRepository
        execute_service = injector.get(TaskExecuteService)  # type: TaskExecuteService

        task_processes = repository.fetch_tasks_by_state(state=TaskState.QUEUED)

        for task in task_processes:
            logger.debug(f'task.key={task.key} is executing')
            execute_service.execute(key=task.key)
            self.check_stop()


class BackgroundWorker:
    _instance = None

    @classmethod
    def get_instance(cls, verbose_log: bool = False):
        if cls._instance is None:
            cls._instance = cls(verbose_log=verbose_log)

        return cls._instance

    def __init__(self, verbose_log: bool = False):
        if verbose_log:
            logger.setLevel(level=logging.INFO)
        else:
            logger.setLevel(level=logging.DEBUG)

        self._noop_worker = NoopTask()
        self._watch_new_task_worker = WatchNewTask()
        self._execute_task_worker = ExecuteTask()

        self._started = False

    def start(self, verbose: bool = False):
        if self._started:
            if verbose:
                logger.debug(f'{self} is already started.')
            return

        self._started = True
        self._noop_worker.start(daemon=False, interval_seconds=100)
        self._watch_new_task_worker.start(daemon=False, interval_seconds=2)
        self._execute_task_worker.start(daemon=False, interval_seconds=3)

    def stop(self):
        if not self._started:
            return

        self._watch_new_task_worker.stop(wait=False)
        self._execute_task_worker.stop(wait=False)

        self._watch_new_task_worker.stop(wait=True)
        self._execute_task_worker.stop(wait=True)
        self._noop_worker.stop(wait=True)

        self._started = False
